#include "RateMatrix_JC.h"

#include <cmath>


/**
 * Constructor of the Jukes-Cantor Rate Matrix class.
 *
 * There are no variables that need to be initialized.
 *
 */
RateMatrix_JC::RateMatrix_JC( void ) 
{
    
}


/**
 * Destructor.
 *
 * Since we do not have any member variables, there is no memory to free here.
 *
 */
RateMatrix_JC::~RateMatrix_JC( void )
{
    
}


/**
 * Calculate the transition probabilities.
 *
 * The transition probabilies are computed as
 *  for off diagonals:      1/4 - 1/4 * exp(-v*4/3)
 *  for diagonals:          1/4 + 3/4 * exp(-v*4/3)
 * Where v = t * r
 *
 * @param   time            The branch time (t) for which to compute the transition probabilities (bl = t * r).
 * @param   rate            The branch rate (r) for which to compute the transition probabilities (bl = t * r).
 * @return                  The transition probability matrix.
 */
std::vector<std::vector<double> > RateMatrix_JC::calculateTransitionProbabilities(double time, double rate) const
{
    
    size_t num_states = 4;
    
    std::vector<std::vector<double> > P = std::vector<std::vector<double> >(num_states, std::vector<double>(num_states, 0.0));
    
    double v = rate * time;
    
    // calculate the transition probabilities
    double bf = 1.0 / num_states;
    double one_minus_bf = 1.0 - bf;
    double p_ii = bf + one_minus_bf * exp(-v/one_minus_bf);
    double p_ij = bf - bf * exp(-v/one_minus_bf);
    for (size_t i=0; i<num_states; i++)
    {
        P[i][i] = p_ii;
        for (size_t j=i+1; j<num_states; j++)
        {
            P[i][j] = p_ij;
            P[j][i] = p_ij;
        }
        
    }
    
    return P;
}
